import csv
import requests
import pkg_resources
import logging

from sgp4.earth_gravity import wgs72
from sgp4.io import twoline2rv
from spacetrack import SpaceTrackClient

logger = logging.getLogger(__name__)

SOURCES_LIST = pkg_resources.resource_filename('satellite_tle', 'sources.csv')

REQUESTS_TIMEOUT = 20  # seconds


def get_tle_sources():
    '''
    Returns a list of (source, url)-tuples for well-known TLE sources.
    '''

    sources = []

    with open(SOURCES_LIST) as csvfile:
        csv_reader = csv.reader(csvfile,
                                delimiter=',',
                                quotechar='\'',
                                quoting=csv.QUOTE_NONNUMERIC)
        for row in csv_reader:
            source, url = row
            sources.append((source, url))

    return sources


def fetch_tle_from_celestrak(norad_cat_id, verify=True):
    '''
    Returns the TLE for a given norad_cat_id as currently available from CelesTrak.
    Raises IndexError if no data is available for the given norad_cat_id.

    Parameters
    ----------
    norad_cat_id : string
        Satellite Catalog Number (5-digit)
    verify : boolean or string, optional
        Either a boolean, in which case it controls whether we verify
        the server's TLS certificate, or a string, in which case it must be a path
        to a CA bundle to use. Defaults to ``True``. (from python-requests)
    '''

    r = requests.get('https://www.celestrak.com/satcat/tle.php?CATNR={}'.format(norad_cat_id),
                     verify=verify,
                     timeout=REQUESTS_TIMEOUT)
    r.raise_for_status()

    if r.text == 'No TLE found':
        raise LookupError

    tle = r.text.split('\r\n')

    return tle[0].strip(), tle[1].strip(), tle[2].strip()


def parse_TLE_file(content):
    '''
    Parses TLE file with 3le format.
    Returns a dictionary of the form {norad_id1: tle1, norad_id2: tle2} for all TLEs found.
    tleN is returned as list of three strings: [satellite_name, line1, line2].
    '''
    tles = dict()
    lines = content.strip().splitlines()

    if len(lines) % 3 != 0:
        raise ValueError

    # Loop over TLEs
    for i in range(len(lines) - 2):
        if (lines[i + 1][0] == "1") & (lines[i + 2][0] == "2"):
            try:
                twoline2rv(lines[i + 1], lines[i + 2], wgs72)
                norad_cat_id = int(lines[i + 1][2:7].encode('ascii'))
                tles[norad_cat_id] = (lines[i].strip(), lines[i + 1], lines[i + 2])
            except ValueError:
                logging.warning('Failed to parse TLE for {}\n({}, {})'.format(
                    lines[i], lines[i + 1], lines[i + 2]))

    return tles


def fetch_tles_from_spacetrack(norad_ids, spacetrack_config):
    '''
    Downloads the TLE set from Space-Track.org.
    Returns a dictionary of the form {norad_id1: tle1, norad_id2: tle2, ...} for all TLEs found.
    tleN is returned as list of three strings: [satellite_name, line1, line2].

    Parameters
    ----------
    norad_ids : set of integers
        Set of Satellite Catalog Numbers (5-digit)
    spacetrack_config : dictionary
        Credentials for log in Space-Track.org following this format:
        {'identity': <username>, 'password': <password>}
    verify : boolean or string, optional
        Either a boolean, in which case it controls whether we verify
        the server's TLS certificate, or a string, in which case it must be a path
        to a CA bundle to use. Defaults to ``True``. (from python-requests)
    '''
    st = SpaceTrackClient(spacetrack_config['identity'], spacetrack_config['password'])
    tles_3le = st.tle_latest(norad_cat_id=norad_ids, ordinal=1, format='3le')

    try:
        return parse_TLE_file(tles_3le)
    except ValueError:
        logging.error('TLE source is malformed.')
        raise ValueError


def fetch_tles_from_url(url, verify=True):
    '''
    Downloads the TLE set from the given url.
    Returns a dictionary of the form {norad_id1: tle1, norad_id2: tle2} for all TLEs found.
    tleN is returned as list of three strings: [satellite_name, line1, line2].

    Parameters
    ----------
    url : string
        URL of the TLE source
    verify : boolean or string, optional
        Either a boolean, in which case it controls whether we verify
        the server's TLS certificate, or a string, in which case it must be a path
        to a CA bundle to use. Defaults to ``True``. (from python-requests)
    '''

    r = requests.get(url, verify=verify, timeout=REQUESTS_TIMEOUT)
    r.raise_for_status()

    try:
        return parse_TLE_file(r.text)
    except ValueError:
        logging.error('TLE source {} is malformed.'.format(url))
        raise ValueError



from setuptools import setup

setup(name='satellitetle',
      version='0.11.1',
      description='Fetch satellite TLEs from various online sources',
      url='https://gitlab.com/librespacefoundation/python-satellitetle',
      author='Fabian P. Schmidt',
      author_email='kerel-fs@gmx.de',
      license='MIT',
      long_description=open('README.rst').read(),
      packages=['satellite_tle'],
      install_requires=[
          'requests~=2.25.0',
          'sgp4~=2.18',
          'spacetrack~=0.16.0',
      ],
      extras_require={'dev': [
          'flake8~=3.9.0',
          'tox~=3.23.0',
      ]},
      package_data={'satellite_tle': ['sources.csv']},
      classifiers=[
          "Programming Language :: Python :: 3",
          "License :: OSI Approved :: MIT License",
      ],
      )



import requests
import logging

from sgp4.earth_gravity import wgs72
from sgp4.io import twoline2rv

from . import (get_tle_sources, fetch_tles_from_url, fetch_tle_from_celestrak,
               fetch_tles_from_spacetrack)


def fetch_all_tles(requested_norad_ids, sources=get_tle_sources(), spacetrack_config=None,
                   verify=True):
    '''
    Returns all TLEs found for the requested satellites available via
    custom TLE sources or via the default sources, Celestrak, CalPoly
    and AMSAT.
    '''
    return fetch_tles(requested_norad_ids, sources, spacetrack_config, verify, only_latest=False)


def fetch_latest_tles(requested_norad_ids, sources=get_tle_sources(), spacetrack_config=None,
                      verify=True):
    '''
    Returns the most recent TLEs found for the requested satellites
    available via custom TLE sources or via the default sources,
    Celestrak, CalPoly and AMSAT.
    '''
    return fetch_tles(requested_norad_ids, sources, spacetrack_config, verify)


def fetch_tles(requested_norad_ids, sources=get_tle_sources(), spacetrack_config=None,
               verify=True, only_latest=True):
    '''
    Returns the most recent or all TLEs found for the requested
    satellites available via custom TLE sources or via the default
    sources, Celestrak, CalPoly and AMSAT.
    '''

    # Dictionary of NORAD IDs with each has as value either a list of
    # 2-tuples of the form (source, tle) or just one 2-tuples
    # depending on only_latest parameter
    # source is a human-readable string
    # tle is a 3-tuple of strings
    tles = dict()

    def update_tles(norad_id, source, tle):
        if norad_id not in requested_norad_ids:
            # Satellite not requested,
            # skip.
            return

        tle_tuple = source, tle

        if norad_id not in tles.keys():
            # Satellite requested and first occurence in the downloaded data,
            # store new TLE.
            if only_latest:
                tles[norad_id] = tle_tuple
            else:
                tles[norad_id] = [tle_tuple]
            return

        if only_latest:
            # There are multiple TLEs for this satellite available.
            # Parse and compare epoch of both TLEs and choose the most recent one.
            current_sat = twoline2rv(tles[norad_id][1][1], tles[norad_id][1][2], wgs72)
            new_sat = twoline2rv(tle[1], tle[2], wgs72)
            if new_sat.epoch > current_sat.epoch:
                # Found a more recent TLE than the current one,
                # store the new TLE.
                logging.debug('Updated {}, epoch '
                              '{:%Y-%m-%d %H:%M:%S} > {:%Y-%m-%d %H:%M:%S}'.format(
                                  norad_id,
                                  new_sat.epoch,
                                  current_sat.epoch))
                tles[norad_id] = tle_tuple
        else:
            tles[norad_id].append(tle_tuple)

    for source, url in sources:
        logging.info('Fetch from {}'.format(url))
        try:
            new_tles = fetch_tles_from_url(url=url, verify=verify)
            logging.debug('Found TLEs for {}'.format(list(new_tles.keys())))
        except (requests.HTTPError, requests.Timeout):
            logging.warning('Failed to download from {}.'.format(source))
            continue
        except ValueError:
            logging.warning('Failed to parse catalog from {}.'.format(source))
            continue

        for norad_id, tle in new_tles.items():
            update_tles(norad_id, source, tle)

    if spacetrack_config:
        try:
            new_tles = fetch_tles_from_spacetrack(requested_norad_ids, spacetrack_config)
            logging.debug('Found TLEs for {}'.format(list(new_tles.keys())))
            for norad_id, tle in new_tles.items():
                update_tles(norad_id, 'Space-Track.org', tle)
        except (requests.HTTPError, requests.Timeout):
            logging.warning('Failed to download from Space-Track.org')
        except ValueError:
            logging.warning('Failed to parse catalog from Space-Track.org')

    # Try fetching missing sats from another Celestrak endoint
    missing_norad_ids = set(requested_norad_ids) - set(tles.keys())

    for norad_id in missing_norad_ids:
        try:
            logging.info('Fetch {} from Celestrak (satcat)'.format(norad_id))
            tle = fetch_tle_from_celestrak(norad_id, verify=verify)
            update_tles(norad_id, 'Celestrak (satcat)', tle)
        except (LookupError, requests.HTTPError, requests.Timeout):
            logging.warning('Fetch {} from Celestrak (satcat) failed!'.format(norad_id))
            continue

    return tles
